class Structure {
	constructor(game, texture, structure, x, y){
		this.game = game;
		
		// Scale
		this.scale = 1;
		this.width = this.scale * texture.width;
		this.height = this.scale * texture.height;
		this.opacity = 1;
		
		// Physics
		this.x = x || 0;
		this.y = y || 0;
		this.r = 0;
		this.vx = 0; // (Math.random() * 2 - 1) * 0.1;
		this.vy = 0; // (Math.random() * 2 - 1) * 0.1;
		this.vr = 0; // (Math.random() * 2 - 1) * 0.2;
		
		this.collision = {
			with: [],
			width: 0,
			height: 0,
			radius: Math.max(this.width, this.height),
			offsetX: 0,
			offsetY: 0,
			
			piercing: false
		};
		
		// Gravity
		this.vxDampG1 = 1.03;
		this.vyDampG1 = 1.03;
		this.vrDampG1 = 1.03;
		
		this.vxDampG0 = 1;
		this.vyDampG0 = 1;
		this.vrDampG0 = 1;
		
		// Structure
		this.maxLife = 30;
		this.life = this.maxLife;
		this.isAlive = true;
		
		this.particleTexture = texture;
		this.texture = CanvasTexture(texture);
		
		this.structure = {
			width: texture.width,
			height: texture.height,
			colors: structure.colors,
			data: JSON.parse(JSON.stringify(structure.data))
		};
		
		this.poll = [];
	}
	
	
	detatch(x, y, childOffsetVX, childOffsetVY){
		var that = this;
		
		// If there is no pixel here, return
		if(this.life<=0 || !this.structure.data[y] || !this.structure.data[y][x]) return false;
		
		if(!this.invincible){
			this.vx += childOffsetVX || 0;
			this.vy += childOffsetVY || 0;
		}
		
		// If this pixel is part of a core, EXPLODE!
		if(!this.invincible && this.isAlive && isCore(this.structure.colors[this.structure.data[y][x]])){
			this.isAlive = false;
			
			// If this structure is not debris, add some cool particles
			if(this.constructor!==Structure){
				for(var i=0; i<this.structure.width; i++){
					for(var j=0; j<this.structure.height; j++){
						if(!this.structure.data[j][i] || Math.random() > 0.2) continue;
						
						var speed = 0.05,
							width = 1,
							height = 1;
						
						var offset = this.game.physics.rotatePoint(
							this.x + this.scale * (i - this.structure.width/2 + width/2),
							this.y + this.scale * (j - this.structure.height/2 + height/2),
							this.r,
							this.x,
							this.y
						);
						
						var offsetVX = speed * (Math.random() * 2 - 1),
							offsetVY = speed * (Math.random() * 2 - 1);
						
						this.game.scene.addParticle(new DebrisParticle(this.game, this.particleTexture, i, j, offset[0], offset[1], this.r, this.vx + offsetVX, this.vy + offsetVY, this.vr));
					}
				}
			}
			
			// Split the boss
			if(this.isBoss){
				for(var i=0; i<4; i++) this.detatch(7 + i, 13, offsetVX/22, offsetVY/22);
				for(var i=0; i<17; i++) this.detatch(15, 25 + i, offsetVX/22, offsetVY/22);
			}
			
			// Split the structure into uneven halfs
			else{
				var x1 = Math.floor(Math.random() * this.structure.width),
					x2 = Math.floor(Math.random() * this.structure.width),
					y1 = Math.floor(Math.random() * this.structure.height),
					y2 = Math.floor(Math.random() * this.structure.height),
					
					xMin = Math.min(x1, x2),
					xMax = this.structure.width - xMin - 1,
					yMin = Math.min(y1, y2),
					yMax = this.structure.height - yMin - 1,
					
					width = xMax - xMin,
					height = yMax - yMin,
					
					step = 1 / Math.max(this.structure.width, this.structure.height),
					index = 0;
				
				while(index < 1){
					var x  = xMin + Math.floor(index * width),
						x_ = xMin + Math.ceil (index * width),
						y  = yMin + Math.floor(index * height),
						y_ = xMin + Math.ceil (index * height),
						x1 = Math.floor(index * this.structure.width),
						y1 = Math.floor(index * this.structure.height);
					
					this.detatch(x, y1, offsetVX/22, offsetVY/22);
					this.detatch(x1, y, offsetVX/22, offsetVY/22);
					
					if(x !== x_) this.detatch(x_, y1, offsetVX/22, offsetVY/22);
					if(y !== y_) this.detatch(x1, y_, offsetVX/22, offsetVY/22);
					
					index += step;
				}
			}
				
			this.life = 0;
			
			for(var y=0; y<this.structure.height; y++){
				for(var x=0; x<this.structure.width; x++){
					if(this.structure.colors[this.structure.data[y][x]] === "156,242,150,1"){
						this.detatch(x, y, 0, 0);
					}
				}
			}
			
			detatchSelf();
		}
		
		
		// This pixel is not a core, detatch it
		else if(!this.invincible) {
			// Create a single particle
			var width = 1,
				height = 1,
				speed = 0.05;
			
			var offset = this.game.physics.rotatePoint(
				this.x + this.scale * (x - this.structure.width/2 + width/2),
				this.y + this.scale * (y - this.structure.height/2 + height/2),
				this.r,
				this.x,
				this.y
			);
			
			var offsetVX = speed * (Math.random() * 2 - 1),
				offsetVY = speed * (Math.random() * 2 - 1);
			
			var particle = new DebrisParticle(this.game, this.particleTexture, x, y, offset[0], offset[1], this.r, this.vx + offsetVX, this.vy + offsetVY, this.vr);
			this.game.scene.addParticle(particle);
			
			
			// Start sorting pixels into groups
			var groups = [],
				groupIndex = -1,
				pixels = [ x + "," + y ],
				structure = this.structure,
				
				maxDepth = Infinity;
			
			var adj = [
				(structure.data[y - 1] || 0)[x - 1] ? 1 : 0,
				(structure.data[y - 1] || 0)[x    ] ? 1 : 0,
				(structure.data[y - 1] || 0)[x + 1] ? 1 : 0,
				(structure.data[y    ] || 0)[x - 1] ? 1 : 0,
				(structure.data[y    ] || 0)[x + 1] ? 1 : 0,
				(structure.data[y + 1] || 0)[x - 1] ? 1 : 0,
				(structure.data[y + 1] || 0)[x    ] ? 1 : 0,
				(structure.data[y + 1] || 0)[x + 1] ? 1 : 0,
			];
			
			var side = adj[1] + adj[3] + adj[4] + adj[6];
			
			if(
				side >= 3 ||
				(
					(adj[1] && adj[4] && !adj[5]) ||
					(adj[4] && adj[6] && !adj[0]) ||
					(adj[6] && adj[3] && !adj[2]) ||
					(adj[3] && adj[1] && !adj[7])
				)
			){
				// Commit the change
				this.texture.context.clearRect(x, y, 1, 1);
				this.structure.data[y][x] = undefined;
				return true;
			}
			
			createGroup(x - 1, y - 1, null, 0);
			createGroup(x    , y - 1, null, 0);
			createGroup(x + 1, y - 1, null, 0);
			createGroup(x - 1, y    , null, 0);
			createGroup(x + 1, y    , null, 0);
			createGroup(x - 1, y + 1, null, 0);
			createGroup(x    , y + 1, null, 0);
			createGroup(x + 1, y + 1, null, 0);
			
			
			// Separate the groups
			for(var i=0; i<groups.length; i++){
				if(i!==groupIndex){
					var group = groups[i],
					
						minX = null,
						minY = null,
						maxX = null,
						maxY = null;
					
					// Detatch into a sizable chunk
					for(var j=0; j<group.length; j++){
						var pos = group[j].split(","),
							px = +pos[0],
							py = +pos[1];
						
						if(j===0){
							minX = px;
							minY = py;
							maxX = px;
							maxY = py;
						}
						
						else{
							minX = Math.min(minX, px);
							minY = Math.min(minY, py);
							maxX = Math.max(maxX, px);
							maxY = Math.max(maxY, py);
						}
					}
					
					var width = maxX - minX + 1,
						height = maxY - minY + 1,
						cloneStructureData = [],
						cloneTexture = CanvasTexture(width, height);
					
					for(var j=0; j<height; j++){
						cloneStructureData[j] = [];
					}
					
					for(var j=0; j<group.length; j++){
						var pos = group[j].split(","),
							px = +pos[0],
							py = +pos[1];
						
						if(!isCore(this.structure.colors[this.structure.data[py][px]])){
							cloneStructureData[py - minY][px - minX] = this.structure.data[py][px];
							cloneTexture.context.drawImage(this.texture, px, py, 1, 1, px - minX, py - minY, 1, 1);
						}
						
						this.texture.context.clearRect(px, py, 1, 1);
						this.structure.data[py][px] = undefined;
					}
					
					var offset = this.game.physics.rotatePoint(
						this.x + this.scale * (minX - this.structure.width/2 + width/2),
						this.y + this.scale * (minY - this.structure.height/2 + height/2),
						this.r,
						this.x,
						this.y
					);
					
					var structureObject = new Structure(this.game, cloneTexture, {
						data: cloneStructureData,
						colors: this.structure.colors,
					}, offset[0], offset[1]);
					
					structureObject.r = this.r;
					structureObject.vr = this.vr + (Math.random() * 2 - 1) * 0.2;
					structureObject.vx = this.vx + (childOffsetVX || 0);
					structureObject.vy = this.vy + (childOffsetVY || 0);
					
					this.game.scene.addObject(structureObject);
				}
			}
			
			// Commit the change
			this.texture.context.clearRect(x, y, 1, 1);
			this.structure.data[y][x] = undefined;
		}
		
		// Recursively assign pixels to a group
		function createGroup(x, y, index, depth){
			if(depth>maxDepth) return index;
			if(that.isBoss && Math.random()>0.5)return index;
			if(!structure.data[y] || !structure.data[y][x]) return index;
			
			var pixel = x + "," + y;
			if(pixels.includes(pixel)) return index;
			
			pixels.push(pixel);
			index = addPixel(pixel, index);
			index = createGroup(x - 1, y - 1, index, depth + 1);
			index = createGroup(x    , y - 1, index, depth + 1);
			index = createGroup(x + 1, y - 1, index, depth + 1);
			index = createGroup(x - 1, y    , index, depth + 1);
			index = createGroup(x + 1, y    , index, depth + 1);
			index = createGroup(x - 1, y + 1, index, depth + 1);
			index = createGroup(x    , y + 1, index, depth + 1);
			index = createGroup(x + 1, y + 1, index, depth + 1);
			
			if(isCore(structure.colors[structure.data[y][x]])){
				groupIndex = index;
			}
			
			return index;
		}
		
		
		// Returns true if color is part of a core
		function isCore(color){
			return color === "156,242,150,1" || color === "195,243,192,1";
		}
		
		
		// Finds the group that pixel belongs to
		function addPixel(pixel, index){
			var group = groups[index];
			
			for(var i=0; i<groups.length; i++){
				if(group !== groups[i] && groups[i].includes(pixel)){
					if(group){
						for(var j=0; j<group.length; j++) groups[i].push(group[j]);
						groups.splice(index, 1);
					}
					
					return i;
				}
			}
			
			if(group){
				group.push(pixel);
				return index;
			}
			
			else{
				groups.push([ pixel ]);
				return groups.length - 1;
			}
		}
		
		
		// Detatch the entire strucure into a new body
		function detatchSelf(){
			var minX = null,
				minY = null,
				maxX = null,
				maxY = null;
			
			for(var y=0; y<that.structure.height; y++){
				for(var x=0; x<that.structure.width; x++){
					if(!that.structure.data[y][x]) continue;
					
					if(minX===null){
						minX = x;
						minY = y;
						maxX = x;
						maxY = y;
					}
					
					else{
						minX = Math.min(minX, x);
						minY = Math.min(minY, y);
						maxX = Math.max(maxX, x);
						maxY = Math.max(maxY, y);
					}
				}
			}
			
			var width = maxX - minX + 1,
				height = maxY - minY + 1,
				cloneStructureData = [],
				cloneTexture = CanvasTexture(width, height);
			
			for(var y=0; y<height; y++){
				cloneStructureData[y] = [];
				
				for(var x=0; x<width; x++){
					var pixel = that.structure.data[minY + y][minX + x];
					
					if(!isCore(that.structure.colors[pixel])){
						cloneStructureData[y][x] = pixel;
						cloneTexture.context.drawImage(that.texture, minX + x, minY + y, 1, 1, x, y, 1, 1);
					}
				}
			}
			
			var offset = that.game.physics.rotatePoint(
				that.x + that.scale * (minX - that.structure.width/2 + width/2),
				that.y + that.scale * (minY - that.structure.height/2 + height/2),
				that.r,
				that.x,
				that.y
			);
			
			var structureObject = new Structure(that.game, cloneTexture, {
				data: cloneStructureData,
				colors: that.structure.colors,
			}, offset[0], offset[1]);
			
			structureObject.r = that.r;
			structureObject.vr = that.vr;
			structureObject.vx = that.vx + (childOffsetVX || 0);
			structureObject.vy = that.vy + (childOffsetVY || 0);
			
			that.game.scene.addObject(structureObject);
		}
		
		return true;
	}
	
	
	update(delta, input){
		// Gravity
		var gravity = this.game.scene.getRule("gravity");
		this.vxDamp = this.vxDampG0 + (this.vxDampG1 - this.vxDampG0) * gravity;
		this.vyDamp = this.vyDampG0 + (this.vyDampG1 - this.vyDampG0) * gravity;
		this.vrDamp = this.vrDampG0 + (this.vrDampG1 - this.vrDampG0) * gravity;
		
		// Physics
		this.x += this.vx * delta * 60;
		this.y += this.vy * delta * 60;
		this.r += this.vr * delta * 60;
		
		this.vx /= this.vxDamp;
		this.vy /= this.vyDamp;
		this.vr /= this.vrDamp;
		
		return this.life -= delta;
	}
	
	
	draw(context, textures, atlas){
		context.save();
		context.translate(this.x, this.y);
		context.rotate(this.r/180*Math.PI);
		
		context.globalAlpha = this.opacity * this.life/this.maxLife * 5;
		context.drawImage(this.texture, -this.width/2, -this.height/2 - 1, this.width, this.height);
		
		if(this.invincible){
			if(this.type===5){
				var width = this.width * 1.15,
					height = this.height * 1.15;
				
				context.drawImage(textures.forcefield, -width/2, -height/2 - 1, width, height);
			}
			
			if(this.type===6){
				var width = this.width * 1.2,
					height = this.height * 1.2;
				
				context.drawImage(textures.forcefield2, -width/2, -height/2 - 1, width, height);
			}
		}
		
		context.restore();
	}
}
